\#    
\#  COPYRIGHT NOTICE
\#    
\#  Copyright (C) 1995, 1996 Michael T. Peterson
\#  This file is part of the PCthreads (tm) multithreading library
\#  package.
\#    
\#  The source files and libraries constituting the PCthreads (tm) package
\#  are free software; you can redistribute them and/or modify them under
\#  the terms of the GNU Library General Public License as published by the Free
\#  Software Foundation; either version 2 of the License, or (at your
\#  option) any later version.
\#    
\#  The PCthreads (tm) package is distributed in the hope that it will
\#  be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
\#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
\#  Library General Public License for more details.
\#    
\#  You should have received a copy of the GNU Library General Public
\#  License along with this library (see the file COPYING.LIB); if not,
\#  write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,
\#  MA 02139, USA.
\#    
\#  To report bugs, bug fixes, or provide feedback you may contact the
\#  author, Michael T. Peterson, at either of the two email addresses listed
\#  below:
\#    
\#             mtp@big.aa.net (preferred)
\#             mtp@zso.dec.com
\#    
\#    
\#  Michael T. Peterson
\#  Redmond, WA.
\#  13 January, 1996
.SH NAME
pthread_cond_timedwait \- Wait for a condition variable to be signaled, or
a time to expire, whichever comes first.
.SH SYNOPSIS
.B #include <pthread.h>
.sp
.B int
.br
.BI "pthread_cond_timedwait(pthread_cond_t *" cond ","
.br
.BI "                       pthread_mutex_t *" mutex ","
.br
.BI "                       const struct timespec *" abs_time )
.SH DESCRIPTION
The
.B pthread_cond_timedwait()
routine causes a thread to wait until a condition variable becomes signaled (or
broadcast), or an expiration time is reached.  Otherwise, this function
behaves exactly like
.B pthread_cond_wait().
.SH PARAMETERS
.SS Input Parameters
.IP cond
The address containing the identifier of the condition variable at which the
calling thread is to wait.  This value may not be NULL.
.IP mutex
The address containing the identifier of the mutex guarding the data predicate.
When 
.B pthread_cond_wait()
is called, the mutex is unlocked and made available to other threads.  When
the condition variable is signaled, the mutex is relocked by the
thread returning from the
.B pthread_cond_wait()
function.
.IP abs_time
The address of a variable containing the expiration time at which the calling
thread will return if the condition variable specified by
.I cond
has not been signaled.
Note that the
.I abs_time
parameter specifies an absolute time, not a time interval.  The structure,
.I struct timespec
may be found in
.I /usr/include/typedefs.h
(furnished with the PCthreads V1.0 Kit).  Its layout is
.sp
.nf
            struct timespec
            {
                long tv_sec;
                long tv_nsec;           /* nanoseconds */
            };
.fi
.sp
To cause a thread to wait on a condition variable until no later than
exactly 12:00:59 AM (59 seconds after midnight), the
.I abs_time
parameter must contain the values
.I abs_time->tv_sec = 59
and
.I abs_time->tv_nsec = 0.
By contrast,
the 
.B pthread_delay_np()
function specifies a relative time (e.g. do not return until N microseconds
have elapsed).  If this same parameter were passed to
.B pthread_delay_np(),
the thread would block for exactly 59 seconds and then return.
.SS Return Parameters
.IP None
.SH RETURN VALUE
A value of zero (0) is returned if no errors are encountered.
When an error condition arises, this routine returns an integer value 
indicating the reason for the error.  Possible error values are:
.TP
.IP EINVAL
The value specified by one or more of the parameters is invalid, or the value of
one or more of the parameters was NULL.
.TP
.IP ENOMEM
The process's heap has become exhausted.
.TP
.IP ENOLCK
The mutex specified by
.I mutex
is not currently locked by the calling thread.
.SH USAGE EXAMPLE
.PP
Lock a mutex and then call 
.B pthread_cond_timedwait().
If the condition variable hasn't been signaled by 1:00 AM, return.
.sp
.nf
                .
                .
                .
    struct timespec ts = {3600, 0};   /* 1:00 AM */


    /*
     * Synchronize child threads with their parent using a condition 
     * variable, of which the predicate (thread_hold) will be controlled
     * by the parent.
     */
    status = pthread_mutex_lock(&cond_mutex);
    CHECK(status, "Failed to lock mutex");

    while( thread_hold )
    {
        status = pthread_cond_timedwait(&cond_var, &cond_mutex, &ts );
        check(status, "2:Cond_wait bad status\n");
    }
                .
                .
            < Access the shared data >
                .
                .
    status = pthread_mutex_unlock(&cond_mutex);
    CHECK(status, "Failed to unlock mutex");

.fi
.PP
Note that the mutex,
.I cond_mutex,
is locked prior to calling
.B pthread_cond_wait().
Because the mutex is reacquired prior to returning from the
.B pthread_cond_wait()
function, the thread may safely access the data guarded by the
predicate
.I thread_hold
boolean.
.SH CONFORMING TO
POSIX .1c
.SH SEE ALSO
.B pthread_mutex_lock(3), pthread_mutex_unlock(3),
.B pthread_cond_signal(3), pthread_cond_broadcast(3),
.B pthread_cond_wait(3)
.SH AUTHOR
Michael T. Peterson <mtp@zso.dec. mtp@big.aa.net>







